//+------------------------------------------------------------------+
//|                                             SetFileConverter.mq5 |
//|                                                  Richard Gunning |
//|                                            https://rjgunning.com |
//+------------------------------------------------------------------+

// Update History
// 1.01 SORA PAPA
//    Mod: The style of this source code is formatted into Google style by ClangFormat.
//    Fix: Changed to skip comment lines.
//    Fix: By not using Split() for key/value extraction, “=” in the value is not removed.
//    Add: Changes the specified key to another value. The key/value pairs are read from [ChangeKeyValue] section in the “SetFileConveter.ini” file.
//    Add: Changes the specified Timeframe key to another value. Timeframe keys are read from [TimeframeKey] section in the “SetFileConveter.ini” file.
//    Add: Changes the specified Applied Price key to another value. Applied Price keys are read from [PriceKey] section in the “SetFileConveter.ini” file.

#define SCRIPT_NAME "SetFileConverter"
#define VERSION "1.01"

#property copyright "Richard Gunning"
#property link "https://rjgunning.com"
#property version VERSION
#property description "Script for converting set files from MT5 format to MT4 format."
#property description "Set files must be in Files Folder or a SubDirectory."
#property script_show_inputs

#include <Files/FileTxt.mqh>

//--- input parameters
input string File         = "";                     // SetFile Name to Convert (Leave blank to convert all)
input string SubDirectory = "";                     // SubDirectory of Files Folder in which to search
input string Directory    = "MT4-Output-Setfiles";  // Output Directory Name (Not Blank)

//--- Create Global Variables
CFileTxt *output = new CFileTxt;
CFileTxt *file   = new CFileTxt;

string KeyValuePairsFile = SCRIPT_NAME + ".ini";

//+------------------------------------------------------------------+
//| MQL4 Compatible Functions                                        |
//+------------------------------------------------------------------+
string StringTrimLeftMQL4(const string __text) {
  string ret_str = __text;
  StringTrimLeft(ret_str);
  return ret_str;
}
#define StringTrimLeft StringTrimLeftMQL4

string StringTrimRightMQL4(const string __text) {
  string ret_str = __text;
  StringTrimRight(ret_str);
  return ret_str;
}
#define StringTrimRight StringTrimRightMQL4
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart() {
  long m_handle;
  bool m_file_found    = true;
  string searchString  = "*.set";
  string thisDirectory = Directory;
  string src_path;
  string dst_path;

  //--- Load key/value pairs
  LoadKeyValuePairs();

  //--- Create Output Directory
  if (thisDirectory == "") {
    thisDirectory = "MT4-Output-Setfiles";
  }
  if (!output.FolderCreate(thisDirectory)) {
    Print("Error Creating Destination Folder");
    delete output;
    delete file;
    return;
  }

  //--- Set Search String and Find First Matching file in Directory
  if (File != "") {
    searchString = File;
  }
  if (SubDirectory == "") {
    m_handle = file.FileFindFirst(searchString, src_path);
  } else {
    m_handle = file.FileFindFirst(SubDirectory + "//" + searchString, src_path);
  }

  //--- Loop through matching files and run conversion
  while (m_handle != INVALID_HANDLE && m_file_found) {
    StringConcatenate(dst_path, thisDirectory, "//", src_path);
    StringReplace(dst_path, ".set", "-MT4.set");
    if (output.Open(dst_path, FILE_WRITE | FILE_REWRITE) != INVALID_HANDLE && file.Open(src_path, FILE_READ) != INVALID_HANDLE) {
      Convert();
    }
    //--- Close SetFiles
    output.Close();
    file.Close();

    //--- Get Next SetFile
    m_file_found = file.FileFindNext(m_handle, src_path);
  }
  //--- Delete Global objects
  delete output;
  delete file;
}
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|  Parse MT5 Set File Line By Line and Write Output Set File       |
//+------------------------------------------------------------------+
void Convert() {
  string line, str;
  string values[2];
  string split[];
  int opt;

  while (!file.IsEnding()) {
    line = file.ReadString();
    if (StringSubstr(line, 0, 1) == ";") {
      output.WriteString(line + "\r\n");
      continue;
    }

    // Read valies[0] and values[1]
    int delimiter_index = StringFind(line, "=");
    if (delimiter_index <= 0) continue;
    values[0] = StringSubstr(line, 0, delimiter_index);
    values[0] = StringTrimRight(StringTrimLeft(values[0]));
    if (values[0] == "") continue;
    values[1] = StringSubstr(line, delimiter_index + 1);
    values[1] = StringTrimRight(StringTrimLeft(values[1]));

    StringSplit(values[1], '|', split);
    if (ArraySize(split) == 0) {
      ArrayResize(split, 1);
    }
    for (int i = 0, count = 0; i < ArraySize(split); i++) {
      if (count == 0) {
        // Changes the specified key to another value
        string change_str = ChangeValue(values[0], split[i]);
        if (change_str != "") {
          str = StringFormat("%s=%s", values[0], change_str);
        } else {
          if (split[i] == NULL) {
            split[i] = "";
          }
          str = StringFormat("%s=%s", values[0], split[i]);
        }
      } else {
        if (split[i] == "" || split[i] == NULL) {
          continue;
        }
        if (count < 4) {
          // Changes the specified key to another value
          string change_str = ChangeValue(values[0], split[i], true);
          if (change_str != "") {
            str = StringFormat("%s,%d=%s", values[0], count, change_str);
          } else {
            if (split[i] == NULL) {
              split[i] = "";
            }
            str = StringFormat("%s,%d=%s", values[0], count, split[i]);
          }
        } else {
          opt = (split[i] == "Y") ? 1 : 0;
          str = StringFormat("%s,F=%d", values[0], opt);
        }
      }
      output.WriteString(str + "\r\n");

      count++;
    }
  }
}
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Changes the specified key to another value                       |
//+------------------------------------------------------------------+
string CHANGE_KEYS[];
string CHANGE_VALUES[];
string TIMEFRAME_KEYS[];
string PRICE_KEYS[];

string ChangeValue(string key, string val = "", bool no_change = false) {
  if (!no_change) {
    for (int i = 0; i < ArraySize(CHANGE_KEYS); i++) {
      if (CHANGE_KEYS[i] == key) {
        return CHANGE_VALUES[i];
      }
    }
  }
  for (int i = 0; i < ArraySize(TIMEFRAME_KEYS); i++) {
    if (TIMEFRAME_KEYS[i] == key) {
      int ival = (int)StringToInteger(val);
      return IntegerToString(TFMigrate(ival));
    }
  }
  for (int i = 0; i < ArraySize(PRICE_KEYS); i++) {
    if (PRICE_KEYS[i] == key) {
      int ival = (int)StringToInteger(val);
      return IntegerToString(PriceMigrate(ival));
    }
  }
  return "";
}
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Load key/value pairs                                             |
//+------------------------------------------------------------------+
#define SECTION_CHEANGE_KEY_VALUE "[ChangeKeyValue]"
#define SECTION_TIMEFRAME_KEY "[TimeframeKey]"
#define SECTION_PRICE_KEY "[PriceKey]"

void LoadKeyValuePairs() {
  int file_handle = FileOpen(KeyValuePairsFile, FILE_TXT | FILE_ANSI | FILE_READ);
  if (file_handle < 0) return;

  int cnt_kv = 0, cnt_tf = 0, cnt_pr = 0;
  int section = 0;
  while (!FileIsEnding(file_handle)) {
    // Read data
    string row           = FileReadString(file_handle);
    string check_comment = StringSubstr(row, 0, 1);
    if (check_comment == "#" || check_comment == ";") continue;

    // Read Section
    if (row == SECTION_CHEANGE_KEY_VALUE) {
      section = 0;
      continue;
    } else if (row == SECTION_TIMEFRAME_KEY) {
      section = 1;
      continue;
    } else if (row == SECTION_PRICE_KEY) {
      section = 2;
      continue;
    }

    int delimiter_index = -1;
    string key = "", val = "";
    switch (section) {
      case 0:  // Read key/value pairs
        delimiter_index = StringFind(row, "=");
        if (delimiter_index <= 0) continue;
        key = StringSubstr(row, 0, delimiter_index);
        key = StringTrimRight(StringTrimLeft(key));
        if (key == "") continue;
        val = StringSubstr(row, delimiter_index + 1);
        val = StringTrimRight(StringTrimLeft(val));
        cnt_kv++;
        // Set key/value pairs
        ArrayResize(CHANGE_KEYS, cnt_kv);
        ArrayResize(CHANGE_VALUES, cnt_kv);
        CHANGE_KEYS[cnt_kv - 1]   = key;
        CHANGE_VALUES[cnt_kv - 1] = val;
        break;

      case 1:  // Read Timeframe keys
        key = StringTrimRight(StringTrimLeft(row));
        cnt_tf++;
        // Set key/value pairs
        ArrayResize(TIMEFRAME_KEYS, cnt_tf);
        TIMEFRAME_KEYS[cnt_tf - 1] = key;
        break;

      case 2:  // Read Price keys
        key = StringTrimRight(StringTrimLeft(row));
        cnt_pr++;
        // Set key/value pairs
        ArrayResize(PRICE_KEYS, cnt_pr);
        PRICE_KEYS[cnt_pr - 1] = key;
        break;

      default:
        break;
    }
  }

  FileClose(file_handle);
}
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Convert timeframe from MQL5 to MQL4                              |
//| https://docs.mql4.com/constants/chartconstants/enum_timeframes   |
//+------------------------------------------------------------------+
int TFMigrate(int __timeframe) {
  switch (__timeframe) {
    case 0:  // PERIOD_CURRENT
      return (0);
    case 1:  // PERIOD_M1
      return (1);
    case 5:  // PERIOD_M5
      return (5);
    case 15:  // PERIOD_M15
      return (15);
    case 30:  // PERIOD_M30
      return (30);
    case 60:  // PERIOD_H1
      return (60);
    case 240:  // PERIOD_H4
      return (240);
    case 1440:  // PERIOD_D1
      return (1440);
    case 10080:  // PERIOD_W1
      return (10080);
    case 43200:  // PERIOD_MN1
      return (43200);

    case 2:  // PERIOD_M2
      return (2);
    case 3:  // PERIOD_M3
      return (3);
    case 4:  // PERIOD_M4
      return (4);
    case 6:  // PERIOD_M6
      return (6);
    case 10:  // PERIOD_M10
      return (10);
    case 12:  // PERIOD_M12
      return (12);
    case 20:  // PERIOD_M20
      return (20);
    case 16385:  // PERIOD_H1
      return (60);
    case 16386:  // PERIOD_H2
      return (120);
    case 16387:  // PERIOD_H3
      return (180);
    case 16388:  // PERIOD_H4
      return (240);
    case 16390:  // PERIOD_H6
      return (360);
    case 16392:  // PERIOD_H8
      return (480);
    case 16396:  // PERIOD_H12
      return (720);
    case 16408:  // PERIOD_D1
      return (1440);
    case 32769:  // PERIOD_W1
      return (10080);
    case 49153:  // PERIOD_MN1
      return (43200);
    default:
      return (0);
  }
}
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//| Convert timeframe from MQL5 to MQL4                              |
//| https://www.mql5.com/en/forum/367341                             |
//+------------------------------------------------------------------+
int PriceMigrate(int __price) {
  switch (__price) {
    case 1:  // PRICE_CLOSE
      return (0);
    case 2:  // PRICE_OPEN
      return (1);
    case 3:  // PRICE_HIGH
      return (2);
    case 4:  // PRICE_LOW
      return (3);
    case 5:  // PRICE_MEDIAN
      return (4);
    case 6:  // PRICE_TYPICAL
      return (5);
    case 7:  // PRICE_WEIGHTED
      return (6);
    default:
      return (0);
  }
}
//+------------------------------------------------------------------+
